#include <MacTypes.h>
#include <FontMgr.h>
#include <WindowMgr.h>
#include <MenuMgr.h>
#include <TextEdit.h>
#include <DialogMgr.h>
#include <EventMgr.h>
#include <DeskMgr.h>
#include <StdfilePkg.h>
#include <FileMgr.h>
#include <ToolboxUtil.h>
#include <ControlMgr.h>
#include <SerialDvr.h>

#include "MBOB.h"

/*
		Created by 	:	Byro
		Date		:	05/14/90
		Modified	:	06/21/90
*/



InitSerial(defaultSettings)
	dlogSettingsPtr	defaultSettings;
	
/*
		Initialize Serial drivers does just that.  Furthermore, it also loads
	the default values into the settings structure.
*/

{
	int		AinRefNumber = AinRefNum, AoutRefNumber = AoutRefNum, Config, Err = noErr;
	Str255	ErrStrng;
	
	
	if ((Err = OpenDriver("\p.Ain", &AinRefNumber)) != noErr)
	{
		/* error dialog and exit to shell */
		
		NumToString((long) Err, ErrStrng);		/* load err val into dlog */
		ParamText(ErrStrng, NIL, NIL, NIL);
		StopAlert(ERR_ALERT, NIL);
		ExitToShell();
	}
		
	if ((Err = OpenDriver("\p.Aout", &AoutRefNumber)) != noErr)
	{
		/* error dialog and exit to shell */
		
		NumToString((long) Err, ErrStrng);		/* load err val into dlog */
		ParamText(ErrStrng, NIL, NIL, NIL);
		StopAlert(ERR_ALERT, NIL);
		ExitToShell();
	}

	Config = baud9600 + data8 + stop20 + noParity;
	
	ResetPorts(Config);		/* reconfigure modem port */
	
	/* load defaults settings structure */
	
	defaultSettings->baud = baud9600;
	defaultSettings->parity = noParity;
	defaultSettings->stopBits = stop20;
	defaultSettings->filter = (int) NIL;
}




shutdownAppl()

/*
		This function closes application by flushing I/O buffer and closing
	the serial drvrs before exiting to the shell.
*/

{
	int		Err = noErr;
	Str255	ErrStrng;
	
	/* flush drvrs */
	
	if ((Err = KillIO(AinRefNum)) != noErr)	/* modem input/output port */
	{
		NumToString((long) Err, ErrStrng);		/* load err val into dlog */
		ParamText(ErrStrng, NIL, NIL, NIL);
		StopAlert(ERR_ALERT, NIL);
	}
	
	if ((Err = KillIO(AoutRefNum)) != noErr)
	{
		NumToString((long) Err, ErrStrng);		/* load err val into dlog */
		ParamText(ErrStrng, NIL, NIL, NIL);
		StopAlert(ERR_ALERT, NIL);
	}
	
	RAMSDClose(sPortA);
	
	ExitToShell();
}




doSerial(Awindow, settings, serStatPtr, filterKountPtr)
	WindowPtr		*Awindow;
	dlogSettingsPtr	settings;
	SerStaRec		*serStatPtr;
	int				*filterKountPtr;
	
/*
		Reads serial input parses status record looking for changes since 
	last iteration.  Additionally, it also parses input buffer looking for
	filter character.
*/

{
	Boolean				invalidRect = FALSE;
	char				serBuf[64];		/* serial input buffer */
	int					i;
	long				bufKount = NIL;	/* I.M. II-178 */
	static SerStaRec	oldSerStat;
	
	/* read buffer once */

	SerGetBuf(AinRefNum, &bufKount);	/* get # of bytes currently in input buffer */
	if ((bufKount < 0) || (bufKount > 63))
		bufKount = NIL;
	if (FSRead(AinRefNum, &bufKount, serBuf) == noErr)
	{
		if (bufKount != NIL)
		{
			for (i = 0; i < bufKount; i++)
			{
				if (serBuf[i] == settings->filter)
					*filterKountPtr += 1;	/* inc filter count */
			}
			SerStatus(AinRefNum, serStatPtr);	/* read status */
			if (cmpStruct(&oldSerStat, serStatPtr, sizeof(oldSerStat)) != TRUE)
				invalidRect = TRUE;
				if ((invalidRect == TRUE) || (*filterKountPtr > 0))
				{
					InvalRect(&(*Awindow)->portRect);	/* generate update evnt */
				}
			oldSerStat = *serStatPtr;
		}
	}
}




displayStatus(SerStatPtr, filterCount)
	SerStaRec	*SerStatPtr;
	int			filterCount;
	
/*
		display all vitals from port status
*/

{
	Byte	SWOverruns, HWOverruns, Parity, Framing;
	int		CursorRow = INITIAL_ROW;		/* current cursor row position */
	
	TextSize(TEXT_FONT_SIZE);
	
	/* mask values */
	
	SWOverruns = swOverrunErr & SerStatPtr->cumErrs;
	HWOverruns = hwOverrunErr & SerStatPtr->cumErrs;
	Parity = parityErr & SerStatPtr->cumErrs;
	Framing = framingErr & SerStatPtr->cumErrs;
	
	/* display status using conditionals coupled with movetos and drawstrings. */
	
	MoveTo(LEFT_MARGIN, CursorRow);
	if (SWOverruns != FALSE)
		DrawString("\pSoftware Overrun error(s) detected   ");
	else
		DrawString("\pno Software Overrun errors detected  ");
	MoveTo(LEFT_MARGIN, CursorRow += ROW_HEIGHT);
	MoveTo(LEFT_MARGIN, CursorRow += ROW_HEIGHT);
	
	if (HWOverruns != FALSE)
		DrawString("\pHardware Overrun error(s) detected   ");
	else
		DrawString("\pno Hardware Overrun errors detected  ");
	MoveTo(LEFT_MARGIN, CursorRow += ROW_HEIGHT);
	MoveTo(LEFT_MARGIN, CursorRow += ROW_HEIGHT);
	
	if (Parity != FALSE)
		DrawString("\pParity error(s) detected   ");
	else
		DrawString("\pno Parity errors detected  ");
	MoveTo(LEFT_MARGIN, CursorRow += ROW_HEIGHT);
	MoveTo(LEFT_MARGIN, CursorRow += ROW_HEIGHT);
	
	if (Framing != FALSE)
		DrawString("\pFraming error(s) detected   ");
	else
		DrawString("\pno Framing errors detected  ");
	MoveTo(LEFT_MARGIN, CursorRow += ROW_HEIGHT);
	MoveTo(LEFT_MARGIN, CursorRow += ROW_HEIGHT);
	
	if (SerStatPtr->ctsHold != FALSE)
		DrawString("\pCTS disabled   ");
	else
		DrawString("\pCTS enabled    ");
	MoveTo(LEFT_MARGIN, CursorRow += ROW_HEIGHT);
	MoveTo(LEFT_MARGIN, CursorRow += ROW_HEIGHT);
	
	if (SerStatPtr->xOffHold != FALSE)
		DrawString("\pStream suspended   ");
	else
		DrawString("\pStream  enabled    ");
	MoveTo(LEFT_MARGIN, CursorRow += ROW_HEIGHT);
	MoveTo(LEFT_MARGIN, CursorRow += ROW_HEIGHT);
	
	if (filterCount > 0)
		DrawString("\pfilter Character detected   ");
	else
		DrawString("\p                            ");	/* don't display */
}




ChangeSettings(settingPtr)
	dlogSettingsPtr    settingPtr;
	
/*
		changes serial port A settings to values in setting structure.
*/

{
	int	Config, Err = noErr;
	
	Config = settingPtr->baud + data8 + settingPtr->parity + settingPtr->stopBits;
	ResetPorts(Config);		/* reconfigure modem port */
}




ResetPorts(config)
	int config;
	
/* resets/reconfigures modem port to value specified by config param */

{
	int 	Err = noErr;
	Str255	ErrStrng;
	
	/* reset modem input serial port */
	
	if ((Err = SerReset(AinRefNum, config)) != noErr)
	{
		/* error dialog and exit to shell */
		
		NumToString((long) Err, ErrStrng);		/* load err val into dlog */
		ParamText(ErrStrng, NIL, NIL, NIL);
		StopAlert(ERR_ALERT, NIL);
		ExitToShell();	
	}
	
	/* reset modem output serial port */
	
	if ((Err = SerReset(AoutRefNum, config)) != noErr)
	{
		/* error dialog and exit to shell */
		
		NumToString((long) Err, ErrStrng);		/* load err val into dlog */
		ParamText(ErrStrng, NIL, NIL, NIL);
		StopAlert(ERR_ALERT, NIL);
		ExitToShell();	
	}
}
